//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	SourceList.cc	--- implementation of methods in SourceList.h
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-12-02: 1.0.0: - first version
//		UL, 1999-12-17: 1.1.0: - brought in line with selectlib-4.23
//		UL, 2000-11-08: 1.2.0: - delete columnData objects in
//								 SourceList::construct()
//
//	AUTHOR:
//		U. Lammers, Thu Dec  2 17:22:32 MET 1999
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/srcdisplay/src/SourceList.cc,v 1.4 2001/03/02 14:49:14 ulammers Exp $
//

#include <cstdio>
#include <iostream>
#include <unistd.h>
#include <cmath>
#include "Tracer.h"
#include "common.h"
#include "errstr.h"
#include "stringutil.h"
#include "DataFilter.h"
#include "SetTableColumn.h"
#include "WCS.h"
#include "SourceList.h"
#include "SasTasks.h"
#include "Ds9.h"
#include "CalOalUtility.h"
#include "Celestial.h"

#include "CalServer.h"
#include "Boresight.h"


double Source::angularSeparation(const Source &s) const
{
	PROC
	RETURN sqrt(CalOalUtility::SQR(_ra - s.ra()) +
				CalOalUtility::SQR(_dec - s.dec())) * RAD_TO_DEG * 3600.;
} // Source::angularSeparation()


SourceList::SourceList(Table *sourceTable, float sourceRadius,
					   bool includeSources)
{
	PROC

	construct(sourceTable, sourceRadius, includeSources);

	ENDPROC
} // SourceList::SourceList()


SourceList::SourceList(const string &boxListSet, float sourceRadius,
					   bool includeSources)
{
	PROC

	DataSet *set = dataSetServer->open(boxListSet, Dal::Read);
	construct(set->table(0), sourceRadius, includeSources);
	dataSetServer->close(set);

	ENDPROC
} // SourceList::SourceList()



void SourceList::construct(Table *table, float sourceRadius,
						   bool includeSources)
{
	PROC

	ColumnData *cdra = table->column("RA")->data(),
			   *cddec = table->column("DEC")->data(),
			   *cdradecerr = table->column("RADEC_ERR")->data(),
			   *cdboxsize = table->column("BOX_SIZE")->data(),
			   *cdcts = table->column("SCTS")->data();
	real32	   *boxsize = cdboxsize->real32Data(),
			   *cts = cdcts->real32Data(),
			   *radecerr = cdradecerr->real32Data();
	real64	   *ra = cdra->real64Data(),
			   *dec = cddec->real64Data();

	for (unsigned n=0; n<table->rows(); ++n) {
		if (cts[n] < 0) {
			errstr<<"Counts for source #"<<n+1<<" is negative: "<<cts[n]
				  <<" - skipping"<<Sas::warning("NegativeSourceCounts");
			continue;
		}

		this->push_back(Source(ra[n]*DEG_TO_RAD, dec[n]*DEG_TO_RAD,
						new FCircle(FPoint(ra[n], dec[n]), sourceRadius,
									includeSources),
						cts[n], boxsize[n], radecerr[n]));
	}

	delete cdra;
	delete cddec;
	delete cdradecerr;
	delete cdboxsize;
	delete cdcts;

	_region.push_back(Component());
	Component   &comp = _region[0];
	unsigned N = this->size();
	for (unsigned n=0; n<N; ++n)
		comp.push_back((*this)[n].shape());

	ENDPROC
} // SourceList::construct()


SourceList::SourceList(const string &eventListName,
					   Xmm::Instrument instrument,
					   const string &xAxis, const string &yAxis,
					   unsigned binX, unsigned binY,
					   float sourceRadius,
					   unsigned nruns, double likemin)
	: _eventListName(eventListName), _instrument(instrument),
	   _xAxis(xAxis), _yAxis(yAxis),
	  _binX(binX), _binY(binY), _sourceRadius(sourceRadius), _nruns(nruns),
	  _likemin(likemin)
{
	PROC

	//	run evselect
	string imageSetName = tmpnam(0);
	SasTasks::evselect.image(eventListName, imageSetName, xAxis, yAxis,
							 binX, binY);

	//	run eboxdetect
	string boxSetName = tmpnam(0);
	SasTasks::eboxdetect.detect(imageSetName, boxSetName, nruns, likemin);

	//	show found sources in ds9
	SasTasks::srcdisplay.show(imageSetName, boxSetName, sourceRadius, false);

    cout<<"\nPlease select the celestial sources that you want to use in the search\n";
	cout<<"for the boresight misalignment angles. For each source please:\n\n";
	cout<<"  1. Double click on the circular marker around the source.\n";
	cout<<"  2. In the dialog box which comes up, change the marker status "
		  "from \"Excluded\"\n     to \"Included\"\n";
	cout<<"  3. In the \"Text\" field specify the true celestial coordinates "
		  "of the source\n";
	cout<<"     in the form:\n";
	cout<<"        23h59m48.12s / 45o38'14.55\"\n";
	cout<<"     or\n";
	cout<<"        123.23445 / 45.238474\n\n";
	cout<<"When you are done, hit any key and the marked sources + coordinates"
		  " will\nbe read\n";

    char c;
    cin.get(c);

	GeoRegion r = Ds9::toRegion(SasTasks::imgdisplay.getRegions());
	for (unsigned i=0; i<r.size(); ++i) {
		Component &c = r[i];
		for (unsigned n=0; n<c.size(); ++n)
			if (c[n]->included())
				//	found a source - see if coordinates are given
				if (c[n]->isLabeled()) {
					vector<string> v(breakup(c[n]->label(), "/#;:"));
					if (v.size() != 2)
						errstr<<"Coordinates for source #"<<n<<" are not "
								"in the right format"
							  <<Sas::error("WrongSourceCoordinates");
					RightAscension ra(v[0]);
					Declination dec(v[1]);
					this->push_back(Source(ra.angle(), dec.angle(),
									c[n], 0, 1, 0.));
				} else
					errstr<<"No coordinates specified for source #"<<n
						  <<Sas::error("NoSourceCoordinates");
	}

	if (this->size() < 2)
		errstr<<"At least 2 sources are needed in the boresight misalignment "
				"angle computation - only "<<this->size()<<" have been "
				"specified."<<Sas::error("TooFewSources");


	//	compute number of counts in source circle for later source cross ID
	//	get WCS info from array and event list
	SetBlock eventList(eventListName);
	WCS	elXwcs(eventList.table(), xAxis), elYwcs(eventList.table(), yAxis);

	SetBlock image(imageSetName);
	WCS imgXwcs(image.array(), 1), imgYwcs(image.array(), 2);
	dataSetServer->close(image.dataSet());

	double minX = elXwcs.refPixel() +
						(imgXwcs(1)-elXwcs.refPixelValue())/elXwcs.delta(),
		   minY = elYwcs.refPixel() +
						(imgYwcs(1)-elYwcs.refPixelValue())/elYwcs.delta();

	for (unsigned n=0; n<this->size(); ++n) {
		DataFilter filter((*this)[n].shape()->
						toSelectExpression(xAxis, yAxis, DBL(binX), DBL(binY),
										   minX, minY));
		vector<DataFilter::RowRange> dum;
		(*this)[n].counts(filter.doSelect(*eventList.table(), dum));
	}		
	dataSetServer->close(eventList.dataSet());

	errstr<<"The following sources will be used in the computation of the\n"
			"boresight misalignment angles:\n"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	errstr<<"Src#        Ra/Dec         counts              shape"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	for (unsigned n=0; n<this->size(); ++n) {
		SourceList &sl = *this;
		RightAscension ra(sl[n].ra());
		Declination dec(sl[n].dec());
		errstr<<n<<' '
			  <<ra<<'/'<<dec<<' '
			  <<sl[n].counts()<<' '
			  <<sl[n].shape()->toSelectExpression(xAxis, yAxis,
									DBL(binX), DBL(binY), minX, minY)<<endl
			  <<Sas::message(Msg::Lib, Msg::Silent);
	}

	unlink(imageSetName.c_str());
	unlink(boxSetName.c_str());

	ENDPROC
} // SourceList::SourceList()



SourceList::SourceList(const string &eventListName,
					   Xmm::Instrument instrument,
					   const string &xAxis, const string &yAxis,
					   const EulerAngles &BSeulerAngles,
					   unsigned binX, unsigned binY,
					   float sourceRadius,
					   unsigned nruns, double likemin,
					   bool runShowDetect)
	: _eventListName(eventListName), _xAxis(xAxis), _yAxis(yAxis),
	  _binX(binX), _binY(binY), _sourceRadius(sourceRadius), _nruns(nruns),
	  _likemin(likemin)
{
	PROC

	calServer.setInstrument(instrument);
	errstr<<"Setting new Boresight Misalignment angles: "
		  <<BSeulerAngles<<Sas::message(Msg::Lib, Msg::Sparse);
	Boresight *bs = calServer.getAtom(bs);
	bs->angles(BSeulerAngles);	

	//	run attcalc
	SasTasks::attcalc.run(eventListName);

	//	run evselect
	string imageSetName = tmpnam(0);
	SasTasks::evselect.image(eventListName, imageSetName, xAxis, yAxis,
							 binX, binY);

	//	run eboxdetect - and show list in ds9
	string boxSetName = tmpnam(0);
	SasTasks::eboxdetect.detect(imageSetName, boxSetName, nruns, likemin);

	//	show found sources in ds9
	if (runShowDetect)
		SasTasks::srcdisplay.show(imageSetName, boxSetName, sourceRadius);

	SetBlock st(boxSetName);
	construct(st.table(), sourceRadius);
	dataSetServer->close(st.dataSet());

	//	compute number of counts in source circle for later source cross ID
	//	get WCS info from array and event list
	SetBlock evl2(eventListName);
	WCS	elXwcs(evl2.table(), xAxis), elYwcs(evl2.table(), yAxis);

	SetBlock image(imageSetName);
	WCS imgXwcs(image.array(), 1), imgYwcs(image.array(), 2);
	dataSetServer->close(image.dataSet());

	double minX = elXwcs.refPixel() +
						(imgXwcs(1)-elXwcs.refPixelValue())/elXwcs.delta(),
		   minY = elYwcs.refPixel() +
						(imgYwcs(1)-elYwcs.refPixelValue())/elYwcs.delta();

	for (unsigned n=0; n<this->size(); ++n) {
		DataFilter filter((*this)[n].shape()->
						toSelectExpression(xAxis, yAxis, DBL(binX), DBL(binY),
										   minX, minY));
		vector<DataFilter::RowRange> dum;
		(*this)[n].counts(filter.doSelect(*evl2.table(), dum));
	}		
	dataSetServer->close(evl2.dataSet());

	errstr<<"Found the following sources for the boresight misalignment "
			"angles: "<<BSeulerAngles<<":\n"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	errstr<<"Src#        Ra/Dec         counts              shape"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	for (unsigned n=0; n<this->size(); ++n) {
		SourceList &sl = *this;
		RightAscension ra(sl[n].ra());
		Declination dec(sl[n].dec());
		errstr<<n<<' '<<ra<<'/'<<dec<<' '
			  <<sl[n].counts()<<' '
			  <<sl[n].shape()->toSelectExpression(xAxis, yAxis,
									DBL(binX), DBL(binY), minX, minY)<<endl
			  <<Sas::message(Msg::Lib, Msg::Silent);
	}
	unlink(imageSetName.c_str());
	unlink(boxSetName.c_str());

	ENDPROC
}  // SourceList::SourceList()


SourceList::~SourceList()
{
	PROC

	_region.free();

	ENDPROC
} // SourceList::~SourceList()


ostream &operator<<(ostream &os, const SourceList &sl)
{
	PROC

	os<<"Source list:\n";
	os<<"   Ra/Dec          counts\n";
	for (unsigned n=0; n<sl.size(); ++n)
		os<<sl[n].ra()*RAD_TO_DEG<<'/'<<sl[n].dec()*RAD_TO_DEG
		  <<' '<<sl[n].counts()<<endl;

	RETURN os;
} // operator<<
