//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	SourceList.h	--- a list of celestial sources
//
//	DESCRIPTION:
//		A SourceList object holds a list of celestial target.
//
//	CHANGE HISTORY:
//		UL, 1999-12-02: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Thu Dec  2 17:22:32 MET 1999
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/srcdisplay/src/SourceList.h,v 1.2 2000/09/14 15:38:47 hsiddiqu Exp $
//

#ifndef SOURCE_LIST_H
#define SOURCE_LIST_H

#include <vector>
#include <string>
#include <iostream>

#include "Dal.h"
#include "Region.h"
#include "Celestial.h"
#include "Xmm.h"

class Source {
public:
	Source(double ra, double dec, FShape *shape, unsigned counts,
		   unsigned boxsize, double raDecErr = 0.)
		: _ra(ra), _dec(dec), _shape(shape), _counts(counts),
		  _boxsize(boxsize), _raDecErr(raDecErr) {}

	double ra() const { return _ra; }
	double dec() const { return _dec; }
	double raDecErr() const { return _raDecErr; }
	FShape *shape() const { return _shape; }
	unsigned counts() const { return _counts; }
	void counts(unsigned nr) { _counts = nr; }
	unsigned boxsize() { return _boxsize; }

	double angularSeparation(const Source &s) const;
		//
		//	PURPOSE:
		//		Return angular separation from other source in arcsecs
		//

	~Source() {}

private:
	double			_ra, _dec;		//	FK5 coordinates
	FShape			*_shape;		//	region around source
	unsigned		_counts,		//	number of counts in
					_boxsize;		//	boxsize
	double			_raDecErr;		//	error associated with _ra/_dec
};


class SourceList : public vector<Source> {
public:
	SourceList() {}
		//
		//	PURPOSE:
		//		Construct empty source list
		//


	SourceList(Table *sourceTable, float sourceRadius = 5,
			   bool includeSources = true);
	SourceList(const string &boxListSet, float sourceRadius = 5,
			   bool includeSources = true);
		//
		//	PURPOSE:
		//		Construct source list from contents of source table
		//		as produced by eboxdetect task
		//
		//	ARGUMENTS:
		//		sourceTable	: pointer to table containing the source list
		//		sourceRadius: will put a circle region around each source
		//					  with this radius in image pixels
		//


	SourceList(const string &eventListName,
			   Xmm::Instrument instrument,
			   const string &xAxis, const string &yAxis,
			   const EulerAngles &BSeulerAngles,
			   unsigned binX = 75, unsigned binY = 75,
			   float sourceRadius = 5, unsigned nruns = 1,
			   double likemin = 15., bool runShowDetect = true);
		//
		//	PURPOSE:
		//		Construct a sourle list from an event list by
		//			a) running attcalc on the event list to compute
		//			   new sky coordinate columns X/Y
		//			b) accumulate an image with evselect
		//			c) run eboxdetect on the image
		//
		//	ARGUMENTS:
		//		eventListName	: name of event list for which to compute
		//						  new sky coordinate columns X/Y
		//		instrument		: instrument ID
		//		xAxis			: name of X axis
		//		yAxis			: name of Y axis
		//		BSeulerAngles	: Euler angles descibing the boresight
		//						  misalignment (3-2-1) according to SIAM ICD
		//		binX/binY		: bin sized in X/Y
		//		sourceRadius	: radius of circle to put around each found
		//						  source
		//		nruns			: number of detection runs in eboxdetect
		//		likemin			: detection likelihood in eboxdetect
		//


	SourceList(const string &eventListName,
			   Xmm::Instrument instrument,
			   const string &xAxis, const string &yAxis,
			   unsigned binX = 75, unsigned binY = 75,
			   float sourceRadius = 5, unsigned nruns = 1,
			   double likemin = 15.);
		//
		//	PURPOSE:
		//		Construct source list from an event list by
		//			a) accumulating an image with evselect
		//			b) running eboxdetect on the image
		//			c) show the image + found sources in Ds9
		//			d) let the user select a number of sources


	const GeoRegion &region() const { return _region; }

	const string &eventListName() const { PROC RETURN _eventListName; }
	Xmm::Instrument instrument() const { PROC RETURN _instrument; }
	const string &xAxis() const { PROC RETURN _xAxis; }
	const string &yAxis() const { PROC RETURN _yAxis; }
	unsigned binX() const { PROC RETURN _binX; }
	unsigned binY() const { PROC RETURN _binY; }
	float sourceRadius() const { PROC RETURN _sourceRadius; }
	unsigned nruns() const { PROC RETURN _nruns; }
	double likemin() const { PROC RETURN _likemin; }

	~SourceList();

private:
	string			_eventListName;
	Xmm::Instrument	_instrument;
	string			_xAxis, _yAxis;
	unsigned		_binX, _binY;
    float           _sourceRadius;
    unsigned          _nruns;
	double			_likemin;
	GeoRegion		_region;
	void construct(Table *sourceTable, float sourceRadius,
				   bool includeSources = true);
};

ostream &operator<<(ostream &os, const SourceList &sl);

#endif


